# The code in this file is copied from https://github.com/praekelt/vumi/blob/master/vumi/codecs/vumi_codecs.py
# which is in turn largely copied from http://stackoverflow.com/questions/13130935/decode-7-bit-gsm
# Vumi's license is included below:

# Copyright (c) Praekelt Foundation and individual contributors.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#     1.  Redistributions of source code must retain the above copyright notice,
#         this list of conditions and the following disclaimer.

#     2.  Redistributions in binary form must reproduce the above copyright
#         notice, this list of conditions and the following disclaimer in the
#         documentation and/or other materials provided with the distribution.

#     3.  Neither the name of the Praekelt Foundation nor the names of its
#         contributors may be used to endorse or promote products derived from
#         this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import sys
import abc
import codecs


# An alternative to using this codec module is to use: https://github.com/dsch/gsm0338
# however, I'm guessing that vumi has been in use longer and we should thus go with it.


class NazCodecException(Exception):
    pass


class GSM7BitCodec(codecs.Codec):
    """
    """

    gsm_basic_charset = (
        "@£$¥èéùìòÇ\nØø\rÅåΔ_ΦΓΛΩΠΨΣΘΞ\x1bÆæßÉ !\"#¤%&'()*+,-./0123456789:;"
        "<=>?¡ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÑÜ`¿abcdefghijklmnopqrstuvwxyzäö"
        "ñüà"
    )

    gsm_basic_charset_map = dict((l, i) for i, l in enumerate(gsm_basic_charset))

    gsm_extension = (
        "````````````````````^```````````````````{}`````\\````````````[~]`"
        "|````````````````````````````````````€``````````````````````````"
    )

    gsm_extension_map = dict((l, i) for i, l in enumerate(gsm_extension))

    def encode(self, input, errors="strict"):
        """
        errors can be 'strict', 'replace' or 'ignore'
        eg:
            xcodec.encode("Zoë","gsm0338") will fail with UnicodeEncodeError
            but
            xcodec.encode("Zoë","gsm0338", 'replace') will return b'Zo?'
            and
            xcodec.encode("Zoë","gsm0338", 'ignore') will return b'Zo'
        """
        result = []
        for position, c in enumerate(input):
            idx = self.gsm_basic_charset_map.get(c)
            if idx is not None:
                result.append(chr(idx))
                continue
            idx = self.gsm_extension_map.get(c)
            if idx is not None:
                result.append(chr(27) + chr(idx))
            else:
                result.append(self.handle_encode_error(c, errors, position, input))

        obj = "".join(result)
        # this is equivalent to;
        # import six; six.b('someString')
        # see:
        # https://github.com/benjaminp/six/blob/68112f3193c7d4bef5ad86ed1b6ed528edd9093d/six.py#L625
        obj = obj.encode("latin-1")
        return (obj, len(obj))

    def handle_encode_error(self, char, handler_type, position, obj):
        handler = getattr(self, "handle_encode_%s_error" % (handler_type,), None)
        if handler is None:
            raise NazCodecException("Invalid errors type %s for GSM7BitCodec", handler_type)
        return handler(char, position, obj)

    @staticmethod
    def handle_encode_strict_error(char, position, obj):
        raise UnicodeEncodeError("gsm0338", char, position, position + 1, repr(obj))

    @staticmethod
    def handle_encode_ignore_error(char, position, obj):
        return ""

    def handle_encode_replace_error(self, char, position, obj):
        return chr(self.gsm_basic_charset_map.get("?"))

    def decode(self, input, errors="strict"):
        """
        errors can be 'strict', 'replace' or 'ignore'
        """
        res = iter(input)
        result = []
        for position, c in enumerate(res):
            try:
                if c == 27:
                    c = next(res)
                    result.append(self.gsm_extension[c])
                else:
                    result.append(self.gsm_basic_charset[c])
            except IndexError as indexErrorException:
                result.append(
                    self.handle_decode_error(c, errors, position, input, indexErrorException)
                )

        obj = "".join(result)
        return (obj, len(obj))

    def handle_decode_error(self, char, handler_type, position, obj, indexErrorException):
        handler = getattr(self, "handle_decode_%s_error" % (handler_type,), None)
        if handler is None:
            raise NazCodecException("Invalid errors type %s for GSM7BitCodec", handler_type)
        return handler(char, position, obj, indexErrorException)

    @staticmethod
    def handle_decode_strict_error(char, position, obj, indexErrorException):
        # https://github.com/google/pytype/issues/349
        raise UnicodeDecodeError(
            "gsm0338",
            chr(char).encode("latin-1"),
            position,
            position + 1,
            repr(obj),  # pytype: disable=wrong-arg-types
        ) from indexErrorException

    @staticmethod
    def handle_decode_ignore_error(char, position, obj, indexErrorException):
        return ""

    @staticmethod
    def handle_decode_replace_error(char, position, obj, indexErrorException):
        return "?"


class UCS2Codec(codecs.Codec):
    """
    UCS2 is for all intents & purposes assumed to be the same as
    big endian UTF16.
    """

    def encode(self, input, errors="strict"):
        # https://github.com/google/pytype/issues/348
        return codecs.utf_16_be_encode(input, errors)  # pytype: disable=module-attr

    def decode(self, input, errors="strict"):
        return codecs.utf_16_be_decode(input, errors)  # pytype: disable=module-attr


class BaseCodec(abc.ABC):
    """
    This is the interface that must be implemented to satisfy naz's encoding/decoding.
    User implementations should inherit this class and
    implement the :func:`__init__ <BaseCodec.__init__>`, :func:`encode <BaseCodec.encode>` and :func:`decode <BaseCodec.decode>`
    methods with the type signatures shown.

    naz calls an implementation of this class to encode/decode messages.
    """

    @abc.abstractmethod
    def __init__(self, encoding: str, errors: str) -> None:
        """
        Parameters:
            encoding: `encoding <https://docs.python.org/3/library/codecs.html#standard-encodings>`_ used to encode messages been sent to SMSC
                      The encoding should be one of the encodings recognised by the SMPP specification. See section 5.2.19 of SMPP spec
                      eg gsm0338, ucs2 etc
            errors:	same meaning as the errors argument to pythons' `encode <https://docs.python.org/3/library/codecs.html#codecs.encode>`_ method
        """
        self.encoding = encoding
        self.errors = errors

    @abc.abstractmethod
    def encode(self, input: str) -> bytes:
        """
        return an encoded version of the string as a bytes object

        Parameters:
            input: the string to encode
        """
        raise NotImplementedError("encode method must be implemented.")

    def decode(self, input: bytes) -> str:
        """
        return a string decoded from the given bytes.

        Parameters:
            input: the bytes to decode
        """
        raise NotImplementedError("decode method must be implemented.")


class SimpleCodec(BaseCodec):
    """
    This is an implementation of `BaseCodec`

    SMPP uses a 7-bit GSM character set. This class implements that encoding/decoding scheme.
    This class can also be used with the usual `python standard encodings <https://docs.python.org/3/library/codecs.html#standard-encodings>`_

    example usage:

    .. highlight:: python
    .. code-block:: python

       ncodec = SimpleCodec(encoding="ucs2")

       ncodec.encode("Zoë")
       ncodec.decode(b'Zo\xc3\xab')
    """

    custom_codecs = {"gsm0338": GSM7BitCodec(), "ucs2": UCS2Codec()}

    def __init__(self, encoding: str = "gsm0338", errors: str = "strict") -> None:
        """
        Parameters:
            encoding: `encoding <https://docs.python.org/3/library/codecs.html#standard-encodings>`_ used to encode messages been sent to SMSC
                      The encoding should be one of the encodings recognised by the SMPP specification. See section 5.2.19 of SMPP spec
            errors:	same meaning as the errors argument to pythons' `encode <https://docs.python.org/3/library/codecs.html#codecs.encode>`_ method
        """
        if not isinstance(encoding, str):
            raise ValueError(
                "`encoding` should be of type:: `str` You entered: {0}".format(type(encoding))
            )
        if not isinstance(errors, str):
            raise ValueError(
                "`errors` should be of type:: `str` You entered: {0}".format(type(errors))
            )
        self.encoding = encoding
        self.errors = errors

    def encode(self, input: str) -> bytes:
        if not isinstance(input, str):
            raise NazCodecException("Only strings accepted for encoding.")
        encoding = self.encoding or sys.getdefaultencoding()
        if encoding in self.custom_codecs:
            encoder = self.custom_codecs[encoding].encode
        else:
            encoder = codecs.getencoder(encoding)

        obj, _ = encoder(input, self.errors)
        return obj

    def decode(self, input: bytes) -> str:
        if not isinstance(input, (bytes, bytearray)):
            raise NazCodecException("Only bytestrings accepted for decoding.")
        encoding = self.encoding or sys.getdefaultencoding()
        if encoding in self.custom_codecs:
            decoder = self.custom_codecs[encoding].decode
        else:
            decoder = codecs.getdecoder(encoding)

        obj, _ = decoder(input, self.errors)
        return obj



import abc
import json

from . import state


NAZ_MESSAGE_PROTOCOL_VERSION = 1
"""
The messages that are published to a queue by either naz
or user application should be versioned.
This version will enable naz to be able to evolve in future;
eg a future version of naz could add/remove the number of required items in a message.
This is a bit similar to: http://docs.celeryproject.org/en/latest/internals/protocol.html
"""


class Message(abc.ABC):
    """
    The message protocol for `naz`. It is the code representation of what
    gets queued into a naz broker.
    This is the interface that must be implemented to satisfy naz's message protocol.

    Users should only ever have to deal with the :class:`SubmitSM <SubmitSM>` implementation
    """

    @abc.abstractmethod
    def __init__(
        self, version: int, smpp_command: str, log_id: str, hook_metadata: str = ""
    ) -> None:
        """
        Parameters:
            version: This indicates the current version of the naz message protocol.
                     This version will enable naz to be able to evolve in future; a future version of `naz` may ship with a different message protocol.
            smpp_command: any one of the SMSC commands eg submit_sm
            log_id: a unique identify of this request
            hook_metadata: a string that to will later on be passed to `naz.Client.hook`. Your application can use it for correlation.
        """
        if not isinstance(version, int):
            raise ValueError(
                "`version` should be of type:: `int` You entered: {0}".format(type(version))
            )
        if version != NAZ_MESSAGE_PROTOCOL_VERSION:
            raise ValueError(
                "`naz` currently only supports naz protocol version {0}".format(
                    NAZ_MESSAGE_PROTOCOL_VERSION
                )
            )
        if not isinstance(smpp_command, str):
            raise ValueError(
                "`smpp_command` should be of type:: `str` You entered: {0}".format(
                    type(smpp_command)
                )
            )
        if not isinstance(log_id, str):
            raise ValueError(
                "`log_id` should be of type:: `str` You entered: {0}".format(type(log_id))
            )
        if not isinstance(hook_metadata, str):
            raise ValueError(
                "`hook_metadata` should be of type:: `str` You entered: {0}".format(
                    type(hook_metadata)
                )
            )
        self.version = version
        self.smpp_command = smpp_command
        self.log_id = log_id
        self.hook_metadata = hook_metadata

    @abc.abstractmethod
    def to_json(self) -> str:
        """
        Serializes the message protocol to json. You can use this method if you would
        like to save the `Message` into a broker like redis/rabbitmq/postgres etc.
        """
        raise NotImplementedError("to_json method must be implemented.")

    @staticmethod
    @abc.abstractmethod
    def from_json(json_message: str) -> "Message":
        """
        Deserializes the message protocol from json.

        Parameters:
            json_message: `naz.protocol.Message` in json format.
        """
        raise NotImplementedError("from_json method must be implemented.")


class SubmitSM(Message):
    """
    The code representation of the `submit_sm` pdu that will get queued into a broker.

    Usage:

    .. highlight:: python
    .. code-block:: python

        import naz

        broker = naz.broker.SimpleBroker(maxsize=1000)
        client = naz.Client(
                smsc_host="127.0.0.1",
                smsc_port=2775,
                system_id="smppclient1",
                password=os.getenv("password", "password"),
                broker=broker,
            )
        msg = naz.protocol.SubmitSM(
            short_message="hello world",
            source_addr="255700111222",
            destination_addr="255799000888",
            log_id="some-id",
        )
        await client.send_message(msg)
    """

    def __init__(
        self,
        short_message: str,
        source_addr: str,
        destination_addr: str,
        log_id: str,
        version: int = 1,
        smpp_command: str = state.SmppCommand.SUBMIT_SM,
        hook_metadata: str = "",
        service_type: str = "CMT",  # section 5.2.11
        source_addr_ton: int = 0x00000001,  # section 5.2.5
        source_addr_npi: int = 0x00000001,
        dest_addr_ton: int = 0x00000001,
        dest_addr_npi: int = 0x00000001,
        # xxxxxx00 store-and-forward
        # xx0010xx Short Message contains ESME Delivery Acknowledgement
        # 00xxxxxx No specific features selected
        esm_class: int = 0b00000011,  # section 5.2.12
        protocol_id: int = 0x00000000,
        priority_flag: int = 0x00000000,
        schedule_delivery_time: str = "",
        validity_period: str = "",
        # xxxxxx01 SMSC Delivery Receipt requested where final delivery outcome is delivery success or failure
        # xxxx01xx SME Delivery Acknowledgement requested
        # xxx0xxxx No Intermediate notification requested
        # all other values reserved
        registered_delivery: int = 0b00000001,  # see section 5.2.17
        replace_if_present_flag: int = 0x00000000,
        sm_default_msg_id: int = 0x00000000,
    ) -> None:
        """
        Parameters:
            short_message: message to send to SMSC
            source_addr: the identifier(eg msisdn) of the message sender
            destination_addr: the identifier(eg msisdn) of the message recipient
            log_id: a unique identify of this request
            version: This indicates the current version of the naz message protocol.
                     This version will enable naz to be able to evolve in future; a future version of `naz` may ship with a different message protocol.
            smpp_command: any one of the SMSC commands eg submit_sm
            hook_metadata: a string that to will later on be passed to `naz.Client.hook`. Your application can use it for correlation.
            service_type:	Indicates the SMS Application service associated with the message
            source_addr_ton:	Type of Number of message originator.
            source_addr_npi:	Numbering Plan Identity of message originator.
            dest_addr_ton:	Type of Number for destination.
            dest_addr_npi:	Numbering Plan Identity of destination
            esm_class:	Indicates Message Mode & Message Type.
            protocol_id:	Protocol Identifier. Network specific field.
            priority_flag:	Designates the priority level of the message.
            schedule_delivery_time:	The short message is to be scheduled by the SMSC for delivery.
            validity_period:	The validity period of this message.
            registered_delivery:	Indicator to signify if an SMSC delivery receipt or an SME acknowledgement is required.
            replace_if_present_flag:	Flag indicating if submitted message should replace an existing message.
            sm_default_msg_id:	Indicates the short message to send from a list of predefined (‘canned’) short messages stored on the SMSC
        """
        self._validate_msg_type_args(
            short_message=short_message,
            source_addr=source_addr,
            destination_addr=destination_addr,
            log_id=log_id,
            version=version,
            hook_metadata=hook_metadata,
            service_type=service_type,
            source_addr_ton=source_addr_ton,
            source_addr_npi=source_addr_npi,
            dest_addr_ton=dest_addr_ton,
            dest_addr_npi=dest_addr_npi,
            esm_class=esm_class,
            protocol_id=protocol_id,
            priority_flag=priority_flag,
            schedule_delivery_time=schedule_delivery_time,
            validity_period=validity_period,
            registered_delivery=registered_delivery,
            replace_if_present_flag=replace_if_present_flag,
            sm_default_msg_id=sm_default_msg_id,
        )

        self.smpp_command: str = state.SmppCommand.SUBMIT_SM
        self.version = version

        self.short_message = short_message
        self.source_addr = source_addr
        self.destination_addr = destination_addr
        self.log_id = log_id
        self.hook_metadata = hook_metadata
        self.service_type = service_type
        self.source_addr_ton = source_addr_ton
        self.source_addr_npi = source_addr_npi
        self.dest_addr_ton = dest_addr_ton
        self.dest_addr_npi = dest_addr_npi
        self.esm_class = esm_class
        self.protocol_id = protocol_id
        self.priority_flag = priority_flag
        self.schedule_delivery_time = schedule_delivery_time
        self.validity_period = validity_period
        self.registered_delivery = registered_delivery
        self.replace_if_present_flag = replace_if_present_flag
        self.sm_default_msg_id = sm_default_msg_id

    @staticmethod
    def _validate_msg_type_args(
        short_message: str,
        source_addr: str,
        destination_addr: str,
        log_id: str,
        version: int,
        hook_metadata: str,
        service_type: str,
        source_addr_ton: int,
        source_addr_npi: int,
        dest_addr_ton: int,
        dest_addr_npi: int,
        esm_class: int,
        protocol_id: int,
        priority_flag: int,
        schedule_delivery_time: str,
        validity_period: str,
        registered_delivery: int,
        replace_if_present_flag: int,
        sm_default_msg_id: int,
    ) -> None:
        if not isinstance(version, int):
            raise ValueError(
                "`version` should be of type:: `int` You entered: {0}".format(type(version))
            )
        if version != NAZ_MESSAGE_PROTOCOL_VERSION:
            raise ValueError(
                "`naz` currently only supports naz protocol version {0}".format(
                    NAZ_MESSAGE_PROTOCOL_VERSION
                )
            )
        if not isinstance(short_message, str):
            raise ValueError(
                "`short_message` should be of type:: `str` You entered: {0}".format(
                    type(short_message)
                )
            )
        if not isinstance(source_addr, str):
            raise ValueError(
                "`source_addr` should be of type:: `str` You entered: {0}".format(type(source_addr))
            )
        if not isinstance(destination_addr, str):
            raise ValueError(
                "`destination_addr` should be of type:: `str` You entered: {0}".format(
                    type(destination_addr)
                )
            )
        if not isinstance(log_id, str):
            raise ValueError(
                "`log_id` should be of type:: `str` You entered: {0}".format(type(log_id))
            )
        if not isinstance(hook_metadata, str):
            raise ValueError(
                "`hook_metadata` should be of type:: `str` You entered: {0}".format(
                    type(hook_metadata)
                )
            )
        if not isinstance(service_type, str):
            raise ValueError(
                "`service_type` should be of type:: `str` You entered: {0}".format(
                    type(service_type)
                )
            )
        if not isinstance(source_addr_ton, int):
            raise ValueError(
                "`source_addr_ton` should be of type:: `int` You entered: {0}".format(
                    type(source_addr_ton)
                )
            )
        if not isinstance(source_addr_npi, int):
            raise ValueError(
                "`source_addr_npi` should be of type:: `int` You entered: {0}".format(
                    type(source_addr_npi)
                )
            )
        if not isinstance(dest_addr_ton, int):
            raise ValueError(
                "`dest_addr_ton` should be of type:: `int` You entered: {0}".format(
                    type(dest_addr_ton)
                )
            )
        if not isinstance(dest_addr_npi, int):
            raise ValueError(
                "`dest_addr_npi` should be of type:: `int` You entered: {0}".format(
                    type(dest_addr_npi)
                )
            )
        if not isinstance(esm_class, int):
            raise ValueError(
                "`esm_class` should be of type:: `int` You entered: {0}".format(type(esm_class))
            )
        if not isinstance(protocol_id, int):
            raise ValueError(
                "`protocol_id` should be of type:: `int` You entered: {0}".format(type(protocol_id))
            )
        if not isinstance(priority_flag, int):
            raise ValueError(
                "`priority_flag` should be of type:: `int` You entered: {0}".format(
                    type(priority_flag)
                )
            )
        if not isinstance(schedule_delivery_time, str):
            raise ValueError(
                "`schedule_delivery_time` should be of type:: `str` You entered: {0}".format(
                    type(schedule_delivery_time)
                )
            )
        if not isinstance(validity_period, str):
            raise ValueError(
                "`validity_period` should be of type:: `str` You entered: {0}".format(
                    type(validity_period)
                )
            )
        if not isinstance(registered_delivery, int):
            raise ValueError(
                "`registered_delivery` should be of type:: `int` You entered: {0}".format(
                    type(registered_delivery)
                )
            )
        if not isinstance(replace_if_present_flag, int):
            raise ValueError(
                "`replace_if_present_flag` should be of type:: `int` You entered: {0}".format(
                    type(replace_if_present_flag)
                )
            )
        if not isinstance(sm_default_msg_id, int):
            raise ValueError(
                "`sm_default_msg_id` should be of type:: `int` You entered: {0}".format(
                    type(sm_default_msg_id)
                )
            )

    def to_json(self) -> str:
        _item = dict(
            smpp_command=self.smpp_command,
            version=self.version,
            short_message=self.short_message,
            source_addr=self.source_addr,
            destination_addr=self.destination_addr,
            log_id=self.log_id,
            hook_metadata=self.hook_metadata,
            service_type=self.service_type,
            source_addr_ton=self.source_addr_ton,
            source_addr_npi=self.source_addr_npi,
            dest_addr_ton=self.dest_addr_ton,
            dest_addr_npi=self.dest_addr_npi,
            esm_class=self.esm_class,
            protocol_id=self.protocol_id,
            priority_flag=self.priority_flag,
            schedule_delivery_time=self.schedule_delivery_time,
            validity_period=self.validity_period,
            registered_delivery=self.registered_delivery,
            replace_if_present_flag=self.replace_if_present_flag,
            sm_default_msg_id=self.sm_default_msg_id,
        )
        return json.dumps(_item)

    @staticmethod
    def from_json(json_message: str) -> "SubmitSM":
        _in_dict = json.loads(json_message)
        return SubmitSM(**_in_dict)


class EnquireLinkResp(Message):
    def __init__(
        self,
        log_id: str,
        sequence_number: int,
        version: int = 1,
        smpp_command: str = state.SmppCommand.ENQUIRE_LINK_RESP,
        hook_metadata: str = "",
    ) -> None:
        """
        Parameters:
            log_id: a unique identify of this request
            version: This indicates the current version of the naz message protocol.
                     This version will enable naz to be able to evolve in future; a future version of `naz` may ship with a different message protocol.
            smpp_command: any one of the SMSC commands eg submit_sm
            hook_metadata: a string that to will later on be passed to `naz.Client.hook`. Your application can use it for correlation.
            sequence_number: SMPP sequence_number
        """
        if not isinstance(log_id, str):
            raise ValueError(
                "`log_id` should be of type:: `str` You entered: {0}".format(type(log_id))
            )
        if not isinstance(sequence_number, int):
            raise ValueError(
                "`sequence_number` should be of type:: `int` You entered: {0}".format(
                    type(sequence_number)
                )
            )
        if not isinstance(version, int):
            raise ValueError(
                "`version` should be of type:: `int` You entered: {0}".format(type(version))
            )
        if version != NAZ_MESSAGE_PROTOCOL_VERSION:
            raise ValueError(
                "`naz` currently only supports naz protocol version {0}".format(
                    NAZ_MESSAGE_PROTOCOL_VERSION
                )
            )
        if not isinstance(smpp_command, str):
            raise ValueError(
                "`smpp_command` should be of type:: `str` You entered: {0}".format(
                    type(smpp_command)
                )
            )
        if not isinstance(hook_metadata, str):
            raise ValueError(
                "`hook_metadata` should be of type:: `str` You entered: {0}".format(
                    type(hook_metadata)
                )
            )
        self.log_id = log_id
        self.sequence_number = sequence_number
        self.version = version
        self.smpp_command = state.SmppCommand.ENQUIRE_LINK_RESP
        self.hook_metadata = hook_metadata

    def to_json(self) -> str:
        _item = dict(
            smpp_command=self.smpp_command,
            version=self.version,
            log_id=self.log_id,
            sequence_number=self.sequence_number,
            hook_metadata=self.hook_metadata,
        )
        return json.dumps(_item)

    @staticmethod
    def from_json(json_message: str) -> "EnquireLinkResp":
        _in_dict = json.loads(json_message)
        return EnquireLinkResp(**_in_dict)


class DeliverSmResp(Message):
    def __init__(
        self,
        log_id: str,
        message_id: str,
        sequence_number: int,
        version: int = 1,
        smpp_command: str = state.SmppCommand.DELIVER_SM_RESP,
        hook_metadata: str = "",
    ) -> None:
        """
        Parameters:
            log_id: a unique identify of this request
            version: This indicates the current version of the naz message protocol.
                     This version will enable naz to be able to evolve in future; a future version of `naz` may ship with a different message protocol.
            smpp_command: any one of the SMSC commands eg submit_sm
            hook_metadata: a string that to will later on be passed to `naz.Client.hook`. Your application can use it for correlation.
            message_id: id of this message
            sequence_number: SMPP sequence_number
        """
        if not isinstance(log_id, str):
            raise ValueError(
                "`log_id` should be of type:: `str` You entered: {0}".format(type(log_id))
            )
        if not isinstance(message_id, str):
            raise ValueError(
                "`message_id` should be of type:: `str` You entered: {0}".format(type(message_id))
            )
        if not isinstance(sequence_number, int):
            raise ValueError(
                "`sequence_number` should be of type:: `int` You entered: {0}".format(
                    type(sequence_number)
                )
            )
        if not isinstance(version, int):
            raise ValueError(
                "`version` should be of type:: `int` You entered: {0}".format(type(version))
            )
        if version != NAZ_MESSAGE_PROTOCOL_VERSION:
            raise ValueError(
                "`naz` currently only supports naz protocol version {0}".format(
                    NAZ_MESSAGE_PROTOCOL_VERSION
                )
            )
        if not isinstance(smpp_command, str):
            raise ValueError(
                "`smpp_command` should be of type:: `str` You entered: {0}".format(
                    type(smpp_command)
                )
            )
        if not isinstance(hook_metadata, str):
            raise ValueError(
                "`hook_metadata` should be of type:: `str` You entered: {0}".format(
                    type(hook_metadata)
                )
            )
        self.log_id = log_id
        self.version = version
        self.smpp_command = state.SmppCommand.DELIVER_SM_RESP
        self.message_id = message_id
        self.sequence_number = sequence_number
        self.hook_metadata = hook_metadata

    def to_json(self) -> str:
        _item = dict(
            smpp_command=self.smpp_command,
            version=self.version,
            log_id=self.log_id,
            message_id=self.message_id,
            sequence_number=self.sequence_number,
            hook_metadata=self.hook_metadata,
        )
        return json.dumps(_item)

    @staticmethod
    def from_json(json_message: str) -> "DeliverSmResp":
        _in_dict = json.loads(json_message)
        return DeliverSmResp(**_in_dict)


def json_to_Message(json_message: str) -> Message:
    """
    Utility function to deserialize the message protocol from json.
    You can use this method if you would like to return the `Message` from a broker like redis/rabbitmq/postgres etc.

    Parameters:
        json_message: `naz.protocol.Message` in json format.
    """
    _item = json.loads(json_message)
    smpp_command = _item["smpp_command"]
    if smpp_command == state.SmppCommand.SUBMIT_SM:
        return SubmitSM.from_json(json_message=json_message)
    elif smpp_command == state.SmppCommand.ENQUIRE_LINK_RESP:
        return EnquireLinkResp.from_json(json_message=json_message)
    elif smpp_command == state.SmppCommand.DELIVER_SM_RESP:
        return DeliverSmResp.from_json(json_message=json_message)
    else:
        raise NotImplementedError(
            "The `from_json` method for smpp_command: `{0}` has not been implemented.".format(
                smpp_command
            )
        )



import abc
import asyncio

from . import protocol


class BaseBroker(abc.ABC):
    """
    This is the interface that must be implemented to satisfy naz's broker.
    User implementations should inherit this class and
    implement the :func:`enqueue <BaseBroker.enqueue>` and :func:`dequeue <BaseBroker.dequeue>` methods with the type signatures shown.

    naz calls an implementation of this class to enqueue and/or dequeue an item.
    """

    @abc.abstractmethod
    async def enqueue(self, message: protocol.Message) -> None:
        """
        enqueue/save an item.

        Parameters:
            message: The item to be enqueued/saved
                  The message is a `naz.protocol.Message` class instance;
                  It is up to the broker implementation to do the serialization(if neccesary) in order to be able to store it.
                  `naz.protocol.Message` has a `to_json()` method that you can use to serialize a `naz.protocol.Message` class instance into json.
        """
        raise NotImplementedError("enqueue method must be implemented.")

    @abc.abstractmethod
    async def dequeue(self) -> protocol.Message:
        """
        dequeue an item.

        Returns:
            item that was dequeued.
            The item has to be returned as a `naz.protocol.Message` class instance.
            It is up to the broker implementation to do the de-serialization(if neccesary).
            `naz.protocol` module has a utility function :func:`json_to_Message <naz.protocol.json_to_Message>`
            that you can use to de-serialize a json string into `naz.protocol.Message` class instance.
        """
        raise NotImplementedError("dequeue method must be implemented.")


class SimpleBroker(BaseBroker):
    """
    This is an in-memory implementation of BaseBroker.

    Note: It should only be used for tests and demo purposes.
    """

    def __init__(self, maxsize: int = 2500) -> None:
        """
        Parameters:
            maxsize: the maximum number of items(not size) that can be put in the queue.
        """
        if not isinstance(maxsize, int):
            raise ValueError(
                "`maxsize` should be of type:: `int` You entered: {0}".format(type(maxsize))
            )
        self.queue: asyncio.queues.Queue = asyncio.Queue(maxsize=maxsize)

    async def enqueue(self, message: protocol.Message) -> None:
        self.queue.put_nowait(message)

    async def dequeue(self) -> protocol.Message:
        return await self.queue.get()



import time
import json
import typing
import logging
import collections
from logging import handlers


class SimpleLogger(logging.Logger):
    """
    It implements a structured logger that renders logs as either json(default) or python dictionary.

    example usage:

    .. highlight:: python
    .. code-block:: python

        logger = SimpleLogger("myLogger")
        logger.log(logging.INFO,
                   {"event": "web_request", "url": "https://www.google.com/"})
    """

    def __init__(
        self,
        logger_name: str,
        level: typing.Union[str, int] = logging.INFO,
        log_metadata: typing.Union[None, dict] = None,
        handler: typing.Union[None, logging.Handler] = None,
        render_as_json=True,
    ) -> None:
        """
        Parameters:
            logger_name: name of the logger. it should be unique per logger.
            level: the level at which to log
            log_metadata: metadata that will be included in all log statements
            handler: python logging `handler <https://docs.python.org/3/library/logging.html#logging.Handler>`_ to be attached to this logger.
                     By default, `logging.StreamHandler` is used.
            render_as_json: whether to render logs as json or dictionary. By default it renders as json.
        """
        super(SimpleLogger, self).__init__(name=logger_name, level=self._nameToLevel(level))
        if not isinstance(logger_name, str):
            raise ValueError(
                "`logger_name` should be of type:: `str` You entered: {0}".format(type(logger_name))
            )
        if not isinstance(level, (int, str)):
            raise ValueError(
                "`level` should be of type:: `str` or `int` You entered: {0}".format(type(level))
            )
        if isinstance(level, str):
            if level.upper() not in ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
                raise ValueError(
                    """`level` should be one of; 'NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'. You entered: {0}""".format(
                        level
                    )
                )
        if not isinstance(log_metadata, (type(None), dict)):
            raise ValueError(
                "`log_metadata` should be of type:: `None` or `dict` You entered: {0}".format(
                    type(log_metadata)
                )
            )
        if not isinstance(handler, (type(None), logging.Handler)):
            raise ValueError(
                "`log_metadata` should be of type:: `None` or `logging.Handler` You entered: {0}".format(
                    type(handler)
                )
            )
        if not isinstance(render_as_json, bool):
            raise ValueError(
                "`render_as_json` should be of type:: `bool` You entered: {0}".format(
                    type(render_as_json)
                )
            )
        self.logger_name = logger_name
        self.level = self._nameToLevel(level)
        if log_metadata is not None:
            self.log_metadata = log_metadata
        else:
            self.log_metadata = {}

        if handler is not None:
            self.handler = handler
        else:
            self.handler = logging.StreamHandler()

        self.render_as_json = render_as_json
        self._set_logger_details()

    def _set_logger_details(self) -> None:
        formatter = logging.Formatter("%(message)s")
        self.handler.setFormatter(formatter)
        self.handler.setLevel(self.level)
        self.addHandler(self.handler)
        self.setLevel(self.level)

    def log(self, level, msg, *args, **kwargs):
        """
        Log 'msg % args' with the integer severity 'level'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.log(level, "We have a %s", "mysterious problem", exc_info=1)
        """
        if self._nameToLevel(level) >= logging.ERROR:
            kwargs.update(dict(exc_info=True))

        new_msg = self._process_msg(msg)
        return super(SimpleLogger, self).log(level, new_msg, *args, **kwargs)

    @staticmethod
    def _nameToLevel(level: typing.Union[str, int]) -> int:
        try:
            if isinstance(level, str):
                # please mypy
                _level: int = logging._nameToLevel[level.upper()]
            else:
                _level = level
            return _level
        except KeyError as e:
            raise ValueError(
                "`level` should be one of; 'NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'"
            ) from e

    def _process_msg(self, msg: typing.Union[str, dict]) -> typing.Union[str, dict]:
        timestamp = self._formatTime()
        if isinstance(msg, dict):
            _timestamp = {"timestamp": timestamp}
            # _timestamp should appear first in resulting dict
            dict_merged_msg = {**_timestamp, **msg, **self.log_metadata}
            if self.render_as_json:
                return self._to_json(dict_merged_msg)
            else:
                return dict_merged_msg
        else:
            str_merged_msg = "{0} {1} {2}".format(timestamp, msg, self.log_metadata)
            if self.log_metadata == {}:
                str_merged_msg = "{0} {1}".format(timestamp, msg)
            if self.render_as_json:
                return self._to_json(str_merged_msg)
            else:
                return str_merged_msg

    def _formatTime(self) -> str:
        """
        Return the creation time of the specified log event as formatted text.

        This code is borrowed from:
        https://docs.python.org/3/library/logging.html#logging.Formatter.formatTime

        The basic behaviour is as follows: an ISO8601-like (or RFC 3339-like) format is used.
        This function uses `time.localtime()` to convert the creation time to a tuple.
        """
        _converter = time.localtime
        _formatter = logging.Formatter()

        now = time.time()
        msecs = (now - int(now)) * 1000

        ct = _converter(now)  # type: ignore
        t = time.strftime(_formatter.default_time_format, ct)
        s = _formatter.default_msec_format % (t, msecs)
        return s

    def _to_json(self, input_msg):
        """
        tries to convert the input message to json and returns it.
        if it fails, it returns the error in string(not json) format
        """
        msg = ""
        try:
            msg = json.dumps(input_msg)
        except Exception as e:
            msg = "naz.SimpleLogger error: {0}".format(str(e))
        return msg


class BreachHandler(handlers.MemoryHandler):
    """
    This is an implementation of `logging.Handler` that puts logs in an in-memory ring buffer.
    When a trigger condition(eg a certain log level) is met;
    then all the logs in the buffer are flushed into a given stream(file, stdout etc)

    It is a bit like
    `logging.handlers.MemoryHandler <https://docs.python.org/3/library/logging.handlers.html#logging.handlers.MemoryHandler>`_
    except that it does not flush when the ring-buffer capacity is met but only when/if the trigger is met.

    It is inspired by the article
    `Triggering Diagnostic Logging on Exception <https://tersesystems.com/blog/2019/07/28/triggering-diagnostic-logging-on-exception/>`_

    example usage:

    .. highlight:: python
    .. code-block:: python

        import naz, logging

        _handler = naz.log.BreachHandler()
        logger = naz.log.SimpleLogger("aha", handler=_handler, log_metadata={"id": "123"})

        logger.log(logging.INFO, {"name": "Jayz"})
        logger.log(logging.ERROR, {"msg": "Houston, we got 99 problems."})

        # or alternatively, to use it with python stdlib logger
        logger = logging.getLogger("my-logger")
        handler = naz.log.BreachHandler()
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        handler.setLevel("DEBUG")
        if not logger.handlers:
            logger.addHandler(handler)
        logger.setLevel("DEBUG")

        logger.info("I did records for Tweet before y'all could even tweet - Dr. Missy Elliot")
        logger.error("damn")
    """

    def __init__(
        self,
        flushLevel: int = logging.WARNING,
        capacity: int = 1_000,
        target: typing.Union[None, logging.Handler] = None,
        flushOnClose: bool = False,
        heartbeatInterval: typing.Union[None, float] = None,
        targetLevel: str = "DEBUG",
    ) -> None:
        """
        Parameters:
            flushLevel: the log level that will trigger this handler to flush logs to :py:attr:`~target`
            capacity: the maximum number of log records to store in the ring buffer
            target: the `log handler <https://docs.python.org/3/library/logging.html#logging.Handler>`_ that will be used.
            flushOnClose: whether to flush the buffer when the handler is closed even if the flush level hasn't been exceeded
            heartbeatInterval: can be a float or None. If it is a float, then a heartbeat log record will be emitted every :py:attr:`~heartbeatInterval` seconds.
                               If it is None(the default), then no heartbeat log record is emitted.
                               If you do decide to set it, a good value is at least 1800(ie 30 minutes).
            targetLevel: the log level to be applied/set to :py:attr:`~target`
        """
        self._validate_args(
            flushLevel=flushLevel,
            capacity=capacity,
            target=target,
            flushOnClose=flushOnClose,
            heartbeatInterval=heartbeatInterval,
            targetLevel=targetLevel,
        )
        if target is None:
            target = logging.StreamHandler()

        # call `logging.handlers.MemoryHandler` init
        super(BreachHandler, self).__init__(  # type: ignore
            capacity=capacity,
            flushLevel=flushLevel,
            target=target,
            flushOnClose=flushOnClose,  # pytype: disable=wrong-keyword-args
        )

        # assuming each log record is 250 bytes, then the maximum
        # memory used by `buffer` will always be == 250*1_000/(1000*1000) == 0.25MB
        self.buffer: typing.Deque[logging.LogRecord] = collections.deque(  # type: ignore
            maxlen=self.capacity  # pytype: disable=attribute-error
        )

        self.heartbeatInterval = heartbeatInterval
        if self.heartbeatInterval:
            self.heartbeatInterval = heartbeatInterval  # seconds
            self._s_time = time.monotonic()

        self.targetLevel: int = logging._nameToLevel[targetLevel.upper()]
        self.target.setLevel(self.targetLevel)  # type: ignore

    def shouldFlush(self, record: logging.LogRecord) -> bool:
        """
        Check for record at the flushLevel or higher.
        Implementation is mostly taken from `logging.handlers.MemoryHandler`
        """
        return record.levelno >= self.flushLevel  # type: ignore # pytype: disable=attribute-error

    def emit(self, record: logging.LogRecord) -> None:
        """
        Emit a record.
        Append the record. If shouldFlush() tells us to, call flush() to process
        the buffer.

        Implementation is taken from `logging.handlers.MemoryHandler`
        """
        self._heartbeat()

        if record.levelno >= self.targetLevel:
            self.buffer.append(record)
        if self.shouldFlush(record):
            self.flush()

    def _heartbeat(self) -> None:
        if not self.heartbeatInterval:
            return

        # check if `heartbeatInterval` seconds have passed.
        # if they have, emit a heartbeat log record to the target handler
        _now = time.monotonic()
        _diff = _now - self._s_time
        if _diff >= self.heartbeatInterval:
            self._s_time = _now
            # see: https://docs.python.org/3/library/logging.html#logging.LogRecord
            record = logging.makeLogRecord(
                {
                    "level": logging.INFO,
                    "name": "BreachHandler",
                    "pathname": ".../naz/naz/log.py",
                    "func": "BreachHandler._heartbeat",
                    "msg": {
                        "event": "naz.BreachHandler.heartbeat",
                        "heartbeatInterval": self.heartbeatInterval,
                    },
                }
            )
            self.target.emit(record=record)  # type: ignore # pytype: disable=attribute-error

    @staticmethod
    def _validate_args(
        flushLevel: int,
        capacity: int,
        target: typing.Union[None, logging.Handler],
        flushOnClose: bool,
        heartbeatInterval: typing.Union[None, float],
        targetLevel: str,
    ) -> None:
        if not isinstance(flushLevel, int):
            raise ValueError(
                "`flushLevel` should be of type:: `int` You entered: {0}".format(type(flushLevel))
            )
        if not isinstance(capacity, int):
            raise ValueError(
                "`capacity` should be of type:: `int` You entered: {0}".format(type(capacity))
            )
        if not isinstance(target, (type(None), logging.Handler)):
            raise ValueError(
                "`target` should be of type:: `None` or `logging.Handler` You entered: {0}".format(
                    type(target)
                )
            )
        if not isinstance(flushOnClose, bool):
            raise ValueError(
                "`flushOnClose` should be of type:: `bool` You entered: {0}".format(
                    type(flushOnClose)
                )
            )
        if not isinstance(heartbeatInterval, (type(None), float)):
            raise ValueError(
                "`heartbeatInterval` should be of type:: `None` or `float` You entered: {0}".format(
                    type(heartbeatInterval)
                )
            )
        if not isinstance(targetLevel, str):
            raise ValueError(
                "`targetLevel` should be of type:: `str` You entered: {0}".format(type(targetLevel))
            )
        if targetLevel.upper() not in ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
            raise ValueError(
                """`targetLevel` should be one of; 'NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'. You entered: {0}""".format(
                    targetLevel
                )
            )
